@@ -55,12 +55,14 @@
 static scm_t_bits screen_tag;
 scm_t_bits window_tag;
 
-SCM equalp_window (SCM x1, SCM x2);
-size_t free_window (SCM x);
-SCM mark_window (SCM x);
-int print_window (SCM x, SCM port, scm_print_state * pstate);
+static size_t gc_free_screen (SCM x);
+static int print_screen (SCM x, SCM port, scm_print_state * pstate);
+
+static SCM equalp_window (SCM x1, SCM x2);
+static size_t gc_free_window (SCM x);
+static SCM mark_window (SCM x);
+static int print_window (SCM x, SCM port, scm_print_state * pstate);
 
-int print_screen (SCM x, SCM port, scm_print_state * pstate);
 
 /* attr -- character attributes, bit flags packed into an unsigned:
    probably uint32 */
@@ -929,16 +931,32 @@ _scm_free_screen (SCM x)
     }
 }
 
-size_t
-free_screen (SCM x)
+/*  This procedure frees a screen without checking its type. In GC
+ *  free procedures, the smob may have already forgotten its type. */
+static size_t
+gc_free_screen (SCM x)
 {
-  assert (SCM_SMOB_PREDICATE (screen_tag, x));
   struct screen_and_ports *sp = (struct screen_and_ports *) SCM_SMOB_DATA (x);
 
   /* Screens should already be null if delwin has been called on them */
   if (sp != NULL)
     {
-      _scm_free_screen (x);
+      if (sp->screen)
+        {
+          delscreen (sp->screen);
+          /* delscreen returns void */
+          sp->screen = NULL;
+        }
+      if (sp->ifp)
+        {
+          fclose (sp->ifp);
+          sp->ifp = NULL;
+        }
+      if (sp->ofp)
+        {
+          fclose (sp->ofp);
+          sp->ofp = NULL;
+        }
       free (sp);
       SCM_SET_SMOB_DATA (x, 0);
     }
@@ -946,17 +964,15 @@ free_screen (SCM x)
   return 0;
 }
 
-
-int
+static int
 print_screen (SCM x, SCM port, scm_print_state * pstate UNUSED)
 {
   struct screen_and_ports *sp;
   SCREEN *screen;
   char str[SIZEOF_VOID_P * 2 + 3];
 
-  /* Don't use _scm_is_screen in this assert, because it says freed screens aren't
-     screens.  */
-  assert (SCM_SMOB_PREDICATE (screen_tag, x));
+  /* A function that may operated on GC'd #<screens>
+   * shouldn't typecheck them, so no typecheck assert here. */
 
   sp = (struct screen_and_ports *) SCM_SMOB_DATA (x);
   screen = sp->screen;
@@ -1055,7 +1071,7 @@ _scm_from_window (WINDOW * x)
 }
 
 // Windows are equal if they point to the same C structures
-SCM
+static SCM
 equalp_window (SCM x1, SCM x2)
 {
   WINDOW *win1, *win2;
@@ -1077,7 +1093,7 @@ equalp_window (SCM x1, SCM x2)
     return SCM_BOOL_T;
 }
 
-SCM
+static SCM
 mark_window (SCM x)
 {
   struct gucu_window *wp;
@@ -1090,14 +1106,12 @@ mark_window (SCM x)
   return SCM_BOOL_F;
 }
 
-size_t
-free_window (SCM x)
+static size_t
+gc_free_window (SCM x)
 {
   struct gucu_window *wp;
   WINDOW *win;
 
-  assert (SCM_SMOB_PREDICATE (window_tag, x));
-
   wp = (struct gucu_window *) SCM_SMOB_DATA (x);
   /* Windows should already be null if delwin has been called on them */
   if (wp != NULL)
@@ -1131,14 +1145,19 @@ free_window (SCM x)
   return 0;
 }
 
-int
+size_t
+free_window (SCM x)
+{
+  assert (SCM_SMOB_PREDICATE (window_tag, x));
+  return gc_free_window (x);
+}
+
+static int
 print_window (SCM _win, SCM port, scm_print_state * pstate UNUSED)
 {
   struct gucu_window *wp = (struct gucu_window *) SCM_SMOB_DATA (_win);
   char str[SIZEOF_VOID_P * 2 + 3];
 
-  assert (SCM_SMOB_PREDICATE (window_tag, _win));
-
   scm_puts ("#<window ", port);
 
   if (wp == (struct gucu_window *) NULL)
@@ -1200,12 +1219,12 @@ gucu_init_type ()
 
       screen_tag = scm_make_smob_type ("screen", sizeof (SCREEN *));
       scm_set_smob_print (screen_tag, print_screen);
-      scm_set_smob_free (screen_tag, free_screen);
+      scm_set_smob_free (screen_tag, gc_free_screen);
       scm_c_define_gsubr ("screen?", 1, 0, 0, gucu_is_screen_p);
 
       window_tag = scm_make_smob_type ("window", sizeof (WINDOW *));
       scm_set_smob_mark (window_tag, mark_window);
-      scm_set_smob_free (window_tag, free_window);
+      scm_set_smob_free (window_tag, gc_free_window);
       scm_set_smob_print (window_tag, print_window);
       scm_set_smob_equalp (window_tag, equalp_window);
       scm_c_define_gsubr ("window?", 1, 0, 0, gucu_is_window_p);
